//#define TEST
//#define OUT
#define outfile "/afs/ece/usr/bre/796/proj/test/stuff.yuv"

//Java native code libraries
#include <jni.h>
#include "nat.h"

//standard libraries
#include <iostream.h>
#include <fstream.h>
#include <math.h>

//our libraries
#include "include/define.h"
#include "include/display.h"
#include "include/dct.h"
#include "include/matrix.h"
#include "include/quantize.h"
#include "include/entropydec.h"
#include "include/huffman.h"

JNIEXPORT jint JNICALL Java_nat_H261decode
(JNIEnv *env, jobject obj, jstring jinfile, jintArray jpels,jintArray jkill){
  //Function that takes an H261 file and decodes and displays it

  //********************************************************
  //********************************************************
  // INITIAL (FIRST FRAME) SETUP FOR DECODING PICTURE
  //********************************************************
  //********************************************************

  // declarations
  int p,comp,m,n,i,j;//loop variables
  int done=FALSE; //assigned to TRUE when last frame is reached
  unsigned char ch;
  pel_t **tmp;
  int dx,dy;
  int size; //used to calculate decoded current frame

  // static declarations
  static int FrameNum=0;
  static int prev=0;//0 or 1, referring to "previous frame". 
                    //(1-prev="current frame")
  static pel_t **C; //C matrix used in DCT and invDCT
  static pel_t **frame[2][3];//holds current & previous frames
  static pel_t **residues[3];//holds residues
  static int **motion_vectors[2];
  static ofstream outstream;

  // Entropy Decoder declarations
  static EntropyDecoder *dec;
  static int *gquant_vals;//GOB quant stepsizes
  static int **mquant_vals;//MB quant stepsizes (-1 if no change from GOB)
  static int *pic_prop;
  static int ***mb_prop;

  //********************************************************
  //Get pointer to filename string from Java side.
  //********************************************************
  const char *infile  = (char *)env->GetStringUTFChars(jinfile,0);

  //********************************************************
  //Get pointer to rgb array from Java side.
  //
  // format of array: each line of the picture (width-wise)
  // is placed one ofter the other.
  //  (w=width of scanline)
  //  index:  0       w-1|w   2*w-1|2*w
  //  array:  ============================...
  //  length: |----------|---------|------...
  //
  // each element of the array is a pixel, represented by an
  // integer in the following format:
  //   0xAARRGGBB
  //      AA = 8-bit Alpha value. Describes how transparent
  //           the pixel is. 0 is completely see-through,
  //           and 255 is completely opaque.
  //      RR = 8-bit Red value
  //      GG = 8-bit Green value
  //      BB = 8-bit Blue value
  //********************************************************
  jint *rgb_pels = (jint *)env->GetIntArrayElements(jpels, 0);

  //********************************************************
  //Get kill signal
  //********************************************************
  jint *k = (jint *)env->GetIntArrayElements(jkill, 0);
  int kill=k[0];
  env->ReleaseIntArrayElements(jkill, k, 0);

  //********************************************************
  // initialize values and data structures
  //********************************************************
  if (0==FrameNum){

    //********************************************************
    //setup Entropy Decoder
    //********************************************************
    dec=new EntropyDecoder(infile,FALSE);
    
#ifdef OUT
    outstream.open(outfile,ios::out);//open file to out
#endif

    //general
    DIM.init(dec->format); //dimensions of each frame

    //********************************************************
    //setup Quantization arrays
    //********************************************************
    if (dec->format==_CIF_){
      gquant_vals=new int[_NUM_CIF_GOB_];
      mquant_vals=new (int *)[_NUM_CIF_GOB_];
      for (i=0;i<12;i++){
	mquant_vals[i]=new int[_NUM_MB_];
      }
    }
    else{
      gquant_vals=new int[_NUM_QCIF_GOB_];
      mquant_vals=new (int *)[_NUM_QCIF_GOB_];
      for (i=0;i<3;i++){
	mquant_vals[i*2]=new int[_NUM_MB_];
      }
    }

    //********************************************************
    // Setup property arrays
    //********************************************************
    pic_prop=new int[_NUM_PIC_PROP_];
    if (dec->format==_CIF_){
      mb_prop=new (int **)[_NUM_CIF_GOB_];
      for (i=0;i<12;i++){
	mb_prop[i]=new (int *)[_NUM_MB_];
	for(j=0;j<_NUM_MB_;j++)
	  mb_prop[i][j]=new int[_NUM_MB_PROP_];
      }
    }
    else{
      mb_prop=new (int **)[_NUM_QCIF_GOB_];
      for (i=0;i<3;i++){
	mb_prop[i*2]=new (int *)[_NUM_MB_];
	for(j=0;j<_NUM_MB_;j++)
	  mb_prop[i*2][j]=new int[_NUM_MB_PROP_];
      }
    }

    //********************************************************
    // Setup C (to be used in DCT operations)
    //********************************************************
    C = new (pel_t *)[_DCT_SIZE_];
    for(i = 0;i < _DCT_SIZE_; i++)
      C[i] = new pel_t[_DCT_SIZE_];
    computeC(C,_DCT_SIZE_);

    //********************************************************
    // Setup frame:
    // 'frame' is a 4-D array, representing the previous and
    //  current picture that is being encoded. referenced as
    //  follows:
    //
    //    frame[{prev,(1-prev)}][{_Y_,_U_,_V_}][row][col]
    //
    //  (note: (1-prev) is used to reference current frame)
    //*******************************************************
    for(p=0;p<2;p++)
      for(comp=_Y_;comp<=_V_;comp++){
	frame[p][comp]=new (pel_t *)[DIM.height[comp]];
	for(m=0;m<DIM.height[comp];m++)
	  frame[p][comp][m]=new pel_t [DIM.width[comp]];
      }//for comp

    //********************************************************
    // Setup residues:
    // 'residues' is a 3-D array, representing the residues
    // of the previous frame before MC, and the current frame
    // after MC. referenced as follows
    //    residues[{_Y_,_U_,_V_}][row][col]
    //*******************************************************
    for(comp=_Y_;comp<=_V_;comp++){
      residues[comp]=new (pel_t *)[DIM.height[comp]];
      for(m=0;m<DIM.height[comp];m++)
	residues[comp][m]=new pel_t [DIM.width[comp]];
    }//for comp
    
    //********************************************************
    // Setup motion_vectors:
    // 'motion_vectors' is a 3-D array, containing the motion
    // vectors that result from doing motion estimation.
    // referenced as follows:
    //    motion_vectors[{_DX_,_DY_}][row][col]
    //
    // - row and col refer to rows and cols of Macroblocks, 
    //   not pels
    // - each motion vector is for all four luminance blocks
    //   in amacroblock.  the motion vector for the other two
    //   color difference blocks (Cb & Cr) is derived by
    //   halving the component values of the macroblock vector
    //   and truncating the magnitude parts toward zero to
    //   yield integer components
    //*******************************************************
    for(m=_DX_;m<=_DY_;m++){
      motion_vectors[m]=new (int *)[DIM.height[_Y_]/_MB_SIZE_];
      for(n=0;n<DIM.height[_Y_]/_MB_SIZE_;n++)
	motion_vectors[m][n]=new int [DIM.width[_Y_]/_MB_SIZE_];
    }//for comp

    //init entropy decoder pointers to data
    dec->init_ptrs(gquant_vals,mquant_vals,pic_prop,mb_prop,
		   &motion_vectors);
    
  }
  //********************************************************
  //********************************************************


  //********************************************************
  //********************************************************
  // DECODE CURRENT FRAME
  //********************************************************
  //********************************************************

  //********************************************************
  // Get decoded bits from file
  //********************************************************

  prev=1-prev;

  //init macroblock properties
  if (dec->format==_CIF_){
    for (i=0;i<12;i++){
      for(j=0;j<_NUM_MB_;j++){
	if (FrameNum==0){
	  mb_prop[i][j][MODE]=INTRA;
	  mb_prop[i][j][MC_PRESENT]=FALSE;
	  mb_prop[i][j][FIL_PRESENT]=FALSE;
	}
	else{
	  mb_prop[i][j][MODE]=INTER;
	  mb_prop[i][j][MC_PRESENT]=FALSE;
	  mb_prop[i][j][FIL_PRESENT]=FALSE;
	}
      }
    }
  }
  else{
    for (i=0;i<3;i++){
      for(j=0;j<_NUM_MB_;j++){
	if (FrameNum==0){
	  mb_prop[i*2][j][MODE]=INTRA;
	  mb_prop[i*2][j][MC_PRESENT]=FALSE;
	  mb_prop[i*2][j][FIL_PRESENT]=FALSE;
	}
	else{
	  mb_prop[i*2][j][MODE]=INTER;
	  mb_prop[i*2][j][MC_PRESENT]=FALSE;
	  mb_prop[i*2][j][FIL_PRESENT]=FALSE;
	}
      }
    }
  }
  

  //init mquant array to -1
  if (dec->format==_CIF_){
    for(m=0;m<12;m++){
      for(n=0;n<_NUM_MB_;n++)
	mquant_vals[m][n]=-1;
    }
  }
  else{
    for(m=0;m<3;m++){
      for(n=0;n<_NUM_MB_;n++)
	mquant_vals[m*2][n]=-1;
    }
  }

  if (FrameNum==0){

    //init frame
    for(comp=_Y_;comp<=_V_;comp++)
      for(m=0;m<DIM.height[comp];m++)
        for(n=0;n<DIM.width[comp];n++)
	  frame[1-prev][comp][m][n]=0;

    done=dec->get_frame(&frame[1-prev]);
  }
  else{

    //init residues
    for(comp=_Y_;comp<=_V_;comp++)
      for(m=0;m<DIM.height[comp];m++)
        for(n=0;n<DIM.width[comp];n++)
	  residues[comp][m][n]=0;

    if (kill)
      done=TRUE;
    else
      done=dec->get_frame(&residues);

  }

  if(kill==0){
    
    //********************************************************
    // invQuantization
    //********************************************************
    for (comp=_Y_;comp<=_V_;comp++)
      if (0==FrameNum)
        DeQuantize(&frame[1-prev][comp],&frame[1-prev][comp],
		   dec->format,comp,gquant_vals,mquant_vals,INTRA);
      else
        DeQuantize(&residues[comp],&residues[comp],dec->format,comp,
	           gquant_vals,mquant_vals,INTER);

    //********************************************************
    // Inverse DCT
    //********************************************************
    tmp = new (pel_t *)[_DCT_SIZE_];
    for(comp=_Y_;comp<=_V_;comp++)
      for(m=0;m<DIM.height[comp];m += _DCT_SIZE_)
	for(n=0;n<DIM.width[comp];n += _DCT_SIZE_){
	  for(i = 0; i < _DCT_SIZE_; i++)
            if (0==FrameNum)
	      tmp[i] = &frame[1-prev][comp][m+i][n];
            else
	      tmp[i] = &residues[comp][m+i][n];
	  invDCT(tmp,C,tmp,_DCT_SIZE_);
	}//for n

    delete tmp;

    //********************************************************
    //make current frame decoded
    //********************************************************

    if (FrameNum>0){ 
      for(m = 0; m < DIM.height[_Y_]; m++){ //Step through MB
	for(n = 0; n < DIM.width[_Y_]; n++){    
	  dx = motion_vectors[_DX_][m/_MB_SIZE_][n/_MB_SIZE_]; //MVs
	  dy = motion_vectors[_DY_][m/_MB_SIZE_][n/_MB_SIZE_];
	  frame[1-prev][_Y_][m][n] = residues[_Y_][m][n]+
	    frame[prev][_Y_][m+dy][n+dx];	  
	}
      }
      for(m = 0; m < DIM.height[_U_]; m++) //Step through MB
	for(n = 0; n < DIM.width[_U_]; n++){    
	  dx = motion_vectors[_DX_][m/_BLOCK_SIZE_][n/_BLOCK_SIZE_]/2; //MVs
	  dy = motion_vectors[_DY_][m/_BLOCK_SIZE_][n/_BLOCK_SIZE_]/2;
	  frame[1-prev][_U_][m][n] = residues[_U_][m][n]+
	    frame[prev][_U_][m+dy][n+dx];
	  frame[1-prev][_V_][m][n] = residues[_V_][m][n]+
	    frame[prev][_V_][m+dy][n+dx];
	}
    }

#ifndef TEST
    
    for(comp=0;comp<3;comp++)
      for(m=0;m<DIM.height[comp];m++)
	for(n=0;n<DIM.width[comp];n++){
	  frame[1-prev][comp][m][n]=clip(frame[1-prev][comp][m][n],0,255);
#ifdef OUT
	  outstream.put((char)frame[1-prev][comp][m][n]);
#endif
	}
    yuv_to_rgb(rgb_pels,&frame[1-prev],DIM);

#else
    
    if (FrameNum==0){
      for(comp=0;comp<3;comp++)
	for(m=0;m<DIM.height[comp];m++)
	  for(n=0;n<DIM.width[comp];n++)
	    frame[1-prev][comp][m][n]=clip(frame[1-prev][comp][m][n],0,255);
      yuv_to_rgb(rgb_pels,&frame[1-prev],DIM);
    }
    else{
      for(comp=0;comp<3;comp++)
	for(m=0;m<DIM.height[comp];m++)
	  for(n=0;n<DIM.width[comp];n++){
	    residues[comp][m][n]=residues[comp][m][n]/2+128;
	    residues[comp][m][n]=clip(residues[comp][m][n],0,255);
	  }
      yuv_to_rgb(rgb_pels,&residues,DIM);
    }
     
#endif
    
  }

  if (done){

    //kill Entropy Decoder
    delete dec;
#ifdef OUT
    outstream.close();
#endif
    //kill quantization arrays
    delete gquant_vals;
    if (dec->format==_CIF_)
      for (i=0;i<12;i++)
	delete mquant_vals[i];
    else
      for (i=0;i<3;i++)
	delete mquant_vals[i*2];
    delete mquant_vals;

    //kill property arrays
    delete pic_prop;
    if (dec->format==_CIF_){
      for (i=0;i<12;i++){
	for(j=0;j<_NUM_MB_;j++)
	  delete mb_prop[i][j];
	delete mb_prop[i];
      }
    }
    else{
      for (i=0;i<3;i++){
	for(j=0;j<_NUM_MB_;j++)
	  delete mb_prop[i*2][j];
	delete mb_prop[i*2];
      }
    }
    delete mb_prop;

    //kill 'frame'
    for(p=0;p<2;p++)
      for(comp=_Y_;comp<_V_;comp++){
	for(m=0;m<DIM.height[comp];m++)
	  delete frame[p][comp][m];
	delete frame[p][comp];
      }//for comp

    //kill 'C'
    for(i = 0;i < _DCT_SIZE_; i++)
      delete C[i];
    delete C;

    //kill 'residues'
    for(comp=_Y_;comp<=_V_;comp++){
      for(m=0;m<DIM.height[comp];m++)
	delete residues[comp][m];
      delete residues[comp];
    }//for comp

    //kill 'motion_vectors'
    for(m=_DX_;m<=_DY_;m++){
      for(n=0;n<DIM.height[_Y_]/_MB_SIZE_;n++)
	delete motion_vectors[m][n];
      delete motion_vectors[m];
    }//for comp
    
    FrameNum=-1;
    
  }//done

  // ********************************************************
  // Release Java variables to allow Java to use the
  // new values assigned to them, and to avoid memory leaks
  //********************************************************
  env->ReleaseStringUTFChars(jinfile, infile);     
  env->ReleaseIntArrayElements(jpels, rgb_pels, 0);

  //********************************************************  
  //********************************************************
  
  FrameNum++;
  return(done);

}//Java_nat_H261decode

JNIEXPORT jint JNICALL Java_nat_H261getFormat
  (JNIEnv *env, jobject obj, jstring jinfile){
  //Function that takes an H261 file and returns its format

  //********************************************************
  //Get pointer to filename string from Java side.
  //********************************************************
  const char *infile  = (char *)env->GetStringUTFChars(jinfile,0);

  //********************************************************
  //Get format info
  //********************************************************
  char buffer[5];
  
  //read in new buffer
  ifstream instream(infile,ios::in);
  instream.read(buffer,4);
  instream.close();

  //********************************************************
  // Release Java variables to allow Java to use the
  // new values assigned to them, and to avoid memory leaks
  //********************************************************
  env->ReleaseStringUTFChars(jinfile, infile);     

  //extract format from bit 29 of file
  return ((buffer[3] & (char)0x08)?_CIF_:_QCIF_);

}//Java_nat_H261getFormat

JNIEXPORT jint JNICALL Java_nat_H261validate
  (JNIEnv *env, jobject obj, jstring jinfile){
  //Function that takes an H261 file and returns 0 if it is valid,
  // 1 otherwise
  
  //********************************************************
  //Get pointer to filename string from Java side.
  //********************************************************
  const char *infile  = (char *)env->GetStringUTFChars(jinfile,0);
  
  //********************************************************
  //Get validation info
  //********************************************************
  char buffer[4];
  
  //read in new buffer
  ifstream instream(infile,ios::in);
  instream.read(buffer,3);
  instream.close();

  if (0!=buffer[0]) return 1;
  if (1!=buffer[1]) return 1;
  if (0!=(buffer[2] & 0xF0)) return 1;   

  //********************************************************
  // Release Java variables to allow Java to use the
  // new values assigned to them, and to avoid memory leaks
  //********************************************************
  env->ReleaseStringUTFChars(jinfile, infile);     

  return 0;

}//Java_nat_H261validate

