//#define TEST

//Java native code libraries
#include <jni.h>
#include "nat.h"

//standard libraries
#include <iostream.h>
#include <fstream.h>
#include <math.h>

//our libraries
#include "include/define.h"
#include "include/display.h"
#include "include/dct.h"
#include "include/matrix.h"
#include "include/quantize.h"
#include "include/filter.h"
#include "include/motionest.h"
#include "include/motioncomp.h"
#include "include/entropyenc.h"

JNIEXPORT jint JNICALL Java_nat_H261encode
(JNIEnv *env, jobject obj, jstring jinfile, jstring joutfile, jintArray jpels,
 jintArray joptions,jintArray jkill){
  //Function that takes in a YUV file (either QCIF or CIF) and encodes it
  //to a H261 file (as well as displaying each encoded frame)
  //
  // infile  = YUV file
  // outfile = H261 file
  // pels    = link to pel array on Java side
  // options: [0] = {_QCIF_,_CIF_}
  //          [1] = {MOTIONEST_FULL}
  //          [2] = quantization threshold
  //          [3] = quantization stepsize

  //********************************************************
  //********************************************************
  // INITIAL (FIRST FRAME) SETUP FOR ENCODING PICTURE
  //********************************************************
  //********************************************************

  // declarations
  int qstepsize,qthreshold,format;
  int p,comp,m,n,i,j;//loop variables
  int done=FALSE; //assigned to TRUE when last frame is reached
  unsigned char ch;
  pel_t **tmp;
  int dx,dy;
  int size; //used to calculate decoded current frame

  // static declarations
  static int FrameNum=0;
  static int prev=0;//0 or 1, referring to "previous frame". 
                    //(1-prev="current frame")
  static pel_t **C; //C matrix used in DCT and invDCT
  static pel_t **frame[2][3];//holds current & previous frames
  static pel_t **residues[3];//holds residues
  static ifstream instream;
  static int **motion_vectors[2];

  // Entropy Encoder declarations
  static EntropyEncoder *enc;
  static int *gquant_vals;//GOB quant stepsizes
  static int **mquant_vals;//MB quant stepsizes (-1 if no change from GOB)
  static int *pic_prop;
  static int ***mb_prop;

  //********************************************************
  //Get pointer to filename strings from Java side.
  //********************************************************
  const char *infile  = (char *)env->GetStringUTFChars(jinfile,0);
  const char *outfile = (char *)env->GetStringUTFChars(joutfile,0);

  //********************************************************
  //Get pointer to rgb array from Java side.
  //
  // format of array: each line of the picture (width-wise)
  // is placed one ofter the other.
  //  (w=width of scanline)
  //  index:  0       w-1|w   2*w-1|2*w
  //  array:  ============================...
  //  length: |----------|---------|------...
  //
  // each element of the array is a pixel, represented by an
  // integer in the following format:
  //   0xAARRGGBB
  //      AA = 8-bit Alpha value. Describes how transparent
  //           the pixel is. 0 is completely see-through,
  //           and 255 is completely opaque.
  //      RR = 8-bit Red value
  //      GG = 8-bit Green value
  //      BB = 8-bit Blue value
  //********************************************************
  jint *rgb_pels = (jint *)env->GetIntArrayElements(jpels, 0);

  //********************************************************
  //Get pointer to options array from Java side.
  //
  // format of array: (see function header)
  //
  //********************************************************
  jint *options = (jint *)env->GetIntArrayElements(joptions, 0);

  //********************************************************
  //Get kill signal
  //********************************************************
  jint *k = (jint *)env->GetIntArrayElements(jkill, 0);
  int kill=k[0];
  env->ReleaseIntArrayElements(jkill, k, 0);

  //init options
  format=options[0];
  qthreshold=options[2];
  qstepsize=options[3];

  //********************************************************
  // initialize values and data structures
  //********************************************************
  if (0==FrameNum){

    //general
    DIM.init(format); //dimensions of each frame

    //********************************************************
    //setup Entropy Encoder
    //********************************************************
    enc=new EntropyEncoder(outfile,FALSE);

    //********************************************************
    //setup Quantization arrays
    //********************************************************
    if (format==_CIF_){
      gquant_vals=new int[_NUM_CIF_GOB_];
      mquant_vals=new (int *)[_NUM_CIF_GOB_];
      for (i=0;i<12;i++)
	mquant_vals[i]=new int[_NUM_MB_];
    }
    else{
      gquant_vals=new int[_NUM_QCIF_GOB_];
      mquant_vals=new (int *)[_NUM_QCIF_GOB_];
      for (i=0;i<3;i++)
	mquant_vals[i*2]=new int[_NUM_MB_];
    }

    //********************************************************
    // Setup property arrays
    //********************************************************
    pic_prop=new int[_NUM_PIC_PROP_];
    if (format==_CIF_){
      mb_prop=new (int **)[_NUM_CIF_GOB_];
      for (i=0;i<12;i++){
	mb_prop[i]=new (int *)[_NUM_MB_];
	for(j=0;j<_NUM_MB_;j++)
	  mb_prop[i][j]=new int[_NUM_MB_PROP_];
      }
    }
    else{
      mb_prop=new (int **)[_NUM_QCIF_GOB_];
      for (i=0;i<3;i++){
	mb_prop[i*2]=new (int *)[_NUM_MB_];
	for(j=0;j<_NUM_MB_;j++)
	  mb_prop[i*2][j]=new int[_NUM_MB_PROP_];
      }
    }

    //********************************************************
    // Setup C (to be used in DCT operations)
    //********************************************************
    C = new (pel_t *)[_DCT_SIZE_];
    for(i = 0;i < _DCT_SIZE_; i++)
      C[i] = new pel_t[_DCT_SIZE_];
    computeC(C,_DCT_SIZE_);

    //********************************************************
    // Setup frame:
    // 'frame' is a 4-D array, representing the previous and
    //  current picture that is being encoded. referenced as
    //  follows:
    //
    //    frame[{prev,(1-prev)}][{_Y_,_U_,_V_}][row][col]
    //
    //  (note: (1-prev) is used to reference current frame)
    //*******************************************************
    for(p=0;p<2;p++)
      for(comp=_Y_;comp<=_V_;comp++){
	frame[p][comp]=new (pel_t *)[DIM.height[comp]];
	for(m=0;m<DIM.height[comp];m++)
	  frame[p][comp][m]=new pel_t [DIM.width[comp]];
      }//for comp

    //********************************************************
    // Setup residues:
    // 'residues' is a 3-D array, representing the residues
    // of the previous frame before MC, and the current frame
    // after MC. referenced as follows
    //    residues[{_Y_,_U_,_V_}][row][col]
    //*******************************************************
    for(comp=_Y_;comp<=_V_;comp++){
      residues[comp]=new (pel_t *)[DIM.height[comp]];
      for(m=0;m<DIM.height[comp];m++)
	residues[comp][m]=new pel_t [DIM.width[comp]];
    }//for comp
    
    //********************************************************
    // Setup motion_vectors:
    // 'motion_vectors' is a 3-D array, containing the motion
    // vectors that result from doing motion estimation.
    // referenced as follows:
    //    motion_vectors[{_DX_,_DY_}][row][col]
    //
    // - row and col refer to rows and cols of Macroblocks, 
    //   not pels
    // - each motion vector is for all four luminance blocks
    //   in amacroblock.  the motion vector for the other two
    //   color difference blocks (Cb & Cr) is derived by
    //   halving the component values of the macroblock vector
    //   and truncating the magnitude parts toward zero to
    //   yield integer components
    //*******************************************************
    for(m=_DX_;m<=_DY_;m++){
      motion_vectors[m]=new (int *)[DIM.height[_Y_]/_MB_SIZE_];
      for(n=0;n<DIM.height[_Y_]/_MB_SIZE_;n++)
	motion_vectors[m][n]=new int [DIM.width[_Y_]/_MB_SIZE_];
    }//for comp

    //*******************************************************
    // get file handles (declared "static" to keep pointers to 
    // the files between each call to this function)
    //*******************************************************
    instream.open(infile,ios::in);//open file to read

  }
  //********************************************************
  //********************************************************


  //********************************************************
  //********************************************************
  // ENCODE CURRENT FRAME
  //********************************************************
  //********************************************************

  //*******************
  // read in next frame
  //*******************
  if (kill)
    done=TRUE;
  else{
    prev=1-prev;
    for(comp=_Y_;comp<=_V_ && done==FALSE;comp++)
      for(m=0;m<DIM.height[comp] && done==FALSE;m++)
	for(n=0;n<DIM.width[comp] && done==FALSE;n++){
	  instream.get(ch);
	  if (instream.eof()) done=TRUE;
	  else
	    frame[1-prev][comp][m][n]=(double)ch;
	}
  }

  if (FrameNum>0 && done==FALSE){
    //********************************************************
    // Motion Estimation
    //********************************************************
    switch(options[1]){//which kind of motion estimation to perform
    case MOTIONEST_FULL:
      MotionEst_Full(motion_vectors,frame,prev,DIM.height,
		     DIM.width,_MB_SIZE_);	      
      break;
    case MOTIONEST_HVMIN:
      MotionEst_HVMin(motion_vectors,frame,prev,DIM.height,
		      DIM.width,_MB_SIZE_);
      break;
    }
    
    /*
	static ifstream instr;
	if (FrameNum==1)
	instr.open("motv_mob.txt",ios::in);
	for(m = 0; m < DIM.height[_Y_]/_MB_SIZE_; m++) //Step through MB
	for(n = 0; n < DIM.width[_Y_]/_MB_SIZE_; n++){
	instr>>motion_vectors[_DX_][m][n]
	>>motion_vectors[_DY_][m][n];
	}
	instr.close();  
	}
    */

			/*
    if (FrameNum!=0){
      //ofstream outstr("motv.txt",ios::out);
      for(m = 0; m < DIM.height[_Y_]/_MB_SIZE_; m++) //Step through MB
	for(n = 0; n < DIM.width[_Y_]/_MB_SIZE_; n++){
	  //outstr<<motion_vectors[_DX_][m][n]<<" "
	  //	  <<motion_vectors[_DY_][m][n]<<endl;
	  motion_vectors[_DX_][m][n]=0;
	  motion_vectors[_DY_][m][n]=0;
	}
      //outstr.close();
    }
			*/

    //********************************************************
    // Motion Compensation
    //********************************************************
    MotionComp(residues,motion_vectors,frame,prev,DIM.height,
	       DIM.width);

    //********************************************************
    // Low Pass Filter
    //********************************************************
    lpf(residues,DIM.height,DIM.width);

  }

  if (done==FALSE){
    //********************************************************
    //DCT Block (assume the DCT coefficient is a size by
    //           size matrix)
    //********************************************************
    tmp = new (pel_t *)[_DCT_SIZE_];
    for(comp=_Y_;comp<=_V_;comp++)
      for(m=0;m<DIM.height[comp];m += _DCT_SIZE_)
	for(n=0;n<DIM.width[comp];n += _DCT_SIZE_){
	  for(i = 0; i < _DCT_SIZE_; i++)
            if (0==FrameNum)
	      tmp[i] = &frame[1-prev][comp][m+i][n];
            else
	      tmp[i] = &residues[comp][m+i][n];
	  DCT(tmp,C,tmp,_DCT_SIZE_);
	}//for n

    //********************************************************
    // Quantization
    //********************************************************
    for (comp=_Y_;comp<=_V_;comp++)
      if (0==FrameNum)
	Quantize(&frame[1-prev][comp],&frame[1-prev][comp],qstepsize,
		 format,comp,gquant_vals,mquant_vals,INTRA,qthreshold);
      else
	Quantize(&residues[comp],&residues[comp],qstepsize,
		 format,comp,gquant_vals,mquant_vals,INTER,qthreshold);

    //********************************************************
    // Send encoded bits to file
    //********************************************************

    //set picture properties
    pic_prop[SPLIT_SCREEN]=OFF;
    pic_prop[DOC_CAM]=OFF;
    pic_prop[FREEZE_PIC]=OFF;
    pic_prop[FORMAT]=format;
    pic_prop[STILL_IMAGE]=1;
    pic_prop[SPARE]=0;

    //set macroblock properties
    if (format==_CIF_){
      for (i=0;i<12;i++){
	for(j=0;j<_NUM_MB_;j++){
	  if (FrameNum==0){
	    mb_prop[i][j][MODE]=INTRA;
	    mb_prop[i][j][MC_PRESENT]=FALSE;
	    mb_prop[i][j][FIL_PRESENT]=FALSE;
	  }
	  else{
	    mb_prop[i][j][MODE]=INTER;
	    mb_prop[i][j][MC_PRESENT]=TRUE;
	    mb_prop[i][j][FIL_PRESENT]=TRUE;
	  }
	}
      }
    }
    else{
      for (i=0;i<3;i++){
	for(j=0;j<_NUM_MB_;j++){
	  if (FrameNum==0){
	    mb_prop[i*2][j][MODE]=INTRA;
	    mb_prop[i*2][j][MC_PRESENT]=FALSE;
	    mb_prop[i*2][j][FIL_PRESENT]=FALSE;
	  }
	  else{
	    mb_prop[i*2][j][MODE]=INTER;
	    mb_prop[i*2][j][MC_PRESENT]=TRUE;
	    mb_prop[i*2][j][FIL_PRESENT]=TRUE;
	  }
	}
      }
    }

    //send next frame
    if (FrameNum==0)
      enc->send_frame(&frame[1-prev],gquant_vals,mquant_vals,pic_prop,
		      mb_prop,&motion_vectors);
    else
      enc->send_frame(&residues,gquant_vals,mquant_vals,pic_prop,
		      mb_prop,&motion_vectors);

    /*    

    cout<<"pic p: "<<pic_prop[0]<<pic_prop[1]<<pic_prop[2]<<pic_prop[3]
	<<pic_prop[4]<<pic_prop[5]<<endl<<endl;
    cout<<"mb_properties: "<<endl;
    //init macroblock properties
    if (format==_CIF_){
      for (i=0;i<12;i++){
	for(j=0;j<_NUM_MB_;j++){
	  cout<<i<<" "<<j<<" "
	      <<mb_prop[i][j][MODE]<<" "
	      <<mb_prop[i][j][MC_PRESENT]<<" "
	      <<mb_prop[i][j][FIL_PRESENT]<<endl;
	}
      }
    }
    else{
      for (i=0;i<3;i++){
	for(j=0;j<_NUM_MB_;j++){
	  cout<<i<<" "<<j<<" "
	      <<mb_prop[i][j][MODE]<<" "
	      <<mb_prop[i][j][MC_PRESENT]<<" "
	      <<mb_prop[i][j][FIL_PRESENT]<<endl;
	}
      }
    }
    
    cout<<endl;

    if (format==_CIF_){
      cout<<"gquant: ";
      for(m=0;m<12;m++)
	cout<<gquant_vals[m]<<" ";
      cout<<endl<<"mquant:"<<endl;
      for(m=0;m<12;m++){
	for(n=0;n<_NUM_MB_;n++)
	  cout<<mquant_vals[m][n]<<" ";
      }
    }
    else{
      cout<<"gquant: ";
      for(m=0;m<3;m++)
	cout<<gquant_vals[m*2]<<" ";
      cout<<endl<<"mquant:"<<endl;
      for(m=0;m<3;m++){
	for(n=0;n<_NUM_MB_;n++)
	  cout<<mquant_vals[m*2][n]<<" ";
      }
    }
    
						     
    if(FrameNum!=0){
      cout<<"motion vectors:"<<endl;
      for(m = 0; m < DIM.height[_Y_]/_MB_SIZE_; m++) //Step through MB
	for(n = 0; n < DIM.width[_Y_]/_MB_SIZE_; n++){
	  cout<<motion_vectors[_DX_][m][n]<<" "
	      <<motion_vectors[_DY_][m][n]<<endl;
	}
    }

    cout<<flush;  

*/

    //********************************************************
    // invQuantization
    //********************************************************
    for (comp=_Y_;comp<=_V_;comp++)
      if (0==FrameNum)
        DeQuantize(&frame[1-prev][comp],&frame[1-prev][comp],
		   format,comp,gquant_vals,mquant_vals,INTRA);
      else
        DeQuantize(&residues[comp],&residues[comp],format,comp,
	           gquant_vals,mquant_vals,INTER);

    //********************************************************
    // Inverse DCT
    //********************************************************
    for(comp=_Y_;comp<=_V_;comp++)
      for(m=0;m<DIM.height[comp];m += _DCT_SIZE_)
	for(n=0;n<DIM.width[comp];n += _DCT_SIZE_){
	  for(i = 0; i < _DCT_SIZE_; i++)
            if (0==FrameNum)
	      tmp[i] = &frame[1-prev][comp][m+i][n];
            else
	      tmp[i] = &residues[comp][m+i][n];
	  invDCT(tmp,C,tmp,_DCT_SIZE_);
	}//for n
    delete tmp;

    //********************************************************
    //make current frame decoded
    //********************************************************

    if (FrameNum>0){ 
      for(m = 0; m < DIM.height[_Y_]; m++){ //Step through MB
	for(n = 0; n < DIM.width[_Y_]; n++){    
	  dx = motion_vectors[_DX_][m/_MB_SIZE_][n/_MB_SIZE_]; //MVs
	  dy = motion_vectors[_DY_][m/_MB_SIZE_][n/_MB_SIZE_];
	  frame[1-prev][_Y_][m][n] = residues[_Y_][m][n]+
	    frame[prev][_Y_][m+dy][n+dx];	  
	}
      }
      for(m = 0; m < DIM.height[_U_]; m++) //Step through MB
	for(n = 0; n < DIM.width[_U_]; n++){    
	  dx = motion_vectors[_DX_][m/_BLOCK_SIZE_][n/_BLOCK_SIZE_]/2; //MVs
	  dy = motion_vectors[_DY_][m/_BLOCK_SIZE_][n/_BLOCK_SIZE_]/2;
	  frame[1-prev][_U_][m][n] = residues[_U_][m][n]+
	    frame[prev][_U_][m+dy][n+dx];
	  frame[1-prev][_V_][m][n] = residues[_V_][m][n]+
	    frame[prev][_V_][m+dy][n+dx];
	}
    }

    /*
    if (FrameNum!=0){
      for(m = 0; m < DIM.height[_Y_]; m++){ //Step through MB
	for(n = 0; n < DIM.width[_Y_]; n++){    
	  cout<<frame[1-prev][_Y_][m][n]<<" ";
	}
	cout<<endl;
      }
      cout<<endl<<endl<<endl<<endl;
    }
    */


#ifndef TEST

    for(comp=0;comp<3;comp++)
      for(m=0;m<DIM.height[comp];m++)
	for(n=0;n<DIM.width[comp];n++)
	  frame[1-prev][comp][m][n]=clip(frame[1-prev][comp][m][n],0,255);
    yuv_to_rgb(rgb_pels,&frame[1-prev],DIM);

#else
    
    if (FrameNum==0){
      for(comp=0;comp<3;comp++)
	for(m=0;m<DIM.height[comp];m++)
	  for(n=0;n<DIM.width[comp];n++)
	    frame[1-prev][comp][m][n]=clip(frame[1-prev][comp][m][n],0,255);
      yuv_to_rgb(rgb_pels,&frame[1-prev],DIM);
    }
    else{
      for(comp=0;comp<3;comp++)
	for(m=0;m<DIM.height[comp];m++)
	  for(n=0;n<DIM.width[comp];n++){
	    residues[comp][m][n]=residues[comp][m][n]/2+128;
	    residues[comp][m][n]=clip(residues[comp][m][n],0,255);
	  }
      yuv_to_rgb(rgb_pels,&residues,DIM);
    }
    
#endif

  }
  else{

    //close file handles
    instream.close();

    //kill Entropy Encoder
    delete enc;

    //kill quantization arrays
    delete gquant_vals;
    if (format==_CIF_)
      for (i=0;i<12;i++)
	delete mquant_vals[i];
    else
      for (i=0;i<3;i++)
	delete mquant_vals[i*2];
    delete mquant_vals;

    //kill property arrays
    delete pic_prop;
    if (format==_CIF_){
      for (i=0;i<12;i++){
	for(j=0;j<_NUM_MB_;j++)
	  delete mb_prop[i][j];
	delete mb_prop[i];
      }
    }
    else{
      for (i=0;i<3;i++){
	for(j=0;j<_NUM_MB_;j++)
	  delete mb_prop[i*2][j];
	delete mb_prop[i*2];
      }
    }
    delete mb_prop;

    //kill 'frame'
    for(p=0;p<2;p++)
      for(comp=_Y_;comp<_V_;comp++){
	for(m=0;m<DIM.height[comp];m++)
	  delete frame[p][comp][m];
	delete frame[p][comp];
      }//for comp

    //kill 'C'
    for(i = 0;i < _DCT_SIZE_; i++)
      delete C[i];
    delete C;

    //kill 'residues'
    for(comp=_Y_;comp<=_V_;comp++){
      for(m=0;m<DIM.height[comp];m++)
	delete residues[comp][m];
      delete residues[comp];
    }//for comp

    //kill 'motion_vectors'
    for(m=_DX_;m<=_DY_;m++){
      for(n=0;n<DIM.height[_Y_]/_MB_SIZE_;n++)
	delete motion_vectors[m][n];
      delete motion_vectors[m];
    }//for comp

    FrameNum=-1;

  }//done

  //********************************************************
  // Release Java variables to allow Java to use the
  // new values assigned to them, and to avoid memory leaks
  //********************************************************
  env->ReleaseStringUTFChars(jinfile, infile);     
  env->ReleaseStringUTFChars(joutfile, outfile); 
  env->ReleaseIntArrayElements(jpels, rgb_pels, 0);
  env->ReleaseIntArrayElements(joptions, options, 0);

  //********************************************************  
  //********************************************************
  
  FrameNum++;
  return(done);

}//Java_gui_H261encode

