//Java native code libraries
#include <jni.h>
#include "nat.h"

//standard libraries
#include <fstream.h>
#include <iostream.h>

//our libraries
#include "include/define.h"
#include "include/display.h"

JNIEXPORT jint JNICALL Java_nat_YUVplay
(JNIEnv *env, jobject obj, jstring jinfile, jintArray jpels, 
 jintArray joptions,jintArray jkill){
  //Function that takes a YUV file and displays it in RGB
  //
  // jinfile = YUV file
  // options: [0] = {_QCIF_,_CIF}

  //********************************************************
  //********************************************************
  // INITIAL (FIRST FRAME) SETUP FOR DECODING PICTURE
  //********************************************************
  //********************************************************

  // declarations
  int comp,m,n;//loop variables
  int done=FALSE; //assigned to TRUE when last frame is reached
  unsigned char ch;

  // static declarations
  static int FrameNum=0;
  static pel_t **frame[3];//holds current & previous frames
  static ifstream instream;

  //********************************************************
  //Get pointer to filename string from Java side.
  //********************************************************
  const char *infile  = (char *)env->GetStringUTFChars(jinfile,0);

  //********************************************************
  //Get pointer to rgb array from Java side.
  //
  // format of array: each line of the picture (width-wise)
  // is placed one ofter the other.
  //  (w=width of scanline)
  //  index:  0       w-1|w   2*w-1|2*w
  //  array:  ============================...
  //  length: |----------|---------|------...
  //
  // each element of the array is a pixel, represented by an
  // integer in the following format:
  //   0xAARRGGBB
  //      AA = 8-bit Alpha value. Describes how transparent
  //           the pixel is. 0 is completely see-through,
  //           and 255 is completely opaque.
  //      RR = 8-bit Red value
  //      GG = 8-bit Green value
  //      BB = 8-bit Blue value
  //********************************************************
  jint *rgb_pels = (jint *)env->GetIntArrayElements(jpels, 0);

  //********************************************************
  //Get pointer to options array from Java side.
  //
  // format of array: (see function header)
  //
  //********************************************************
  jint *options = (jint *)env->GetIntArrayElements(joptions, 0);

  //********************************************************
  //Get kill signal
  //********************************************************
  jint *k = (jint *)env->GetIntArrayElements(jkill, 0);
  int kill=k[0];
  env->ReleaseIntArrayElements(jkill, k, 0);

  //********************************************************
  // initialize values and data structures
  //********************************************************
  if (0==FrameNum){

    //general
    DIM.init(options[0]); //dimensions of each frame

    //********************************************************
    // Setup frame:
    // 'frame' is a 4-D array, representing the previous and
    //  current picture that is being encoded. referenced as
    //  follows:
    //
    //    frame[{prev,(1-prev)}][{_Y_,_U_,_V_}][row][col]
    //
    //  (note: (1-prev) is used to reference current frame)
    //*******************************************************
    for(comp=_Y_;comp<=_V_;comp++){
      frame[comp]=new (pel_t *)[DIM.height[comp]];
      for(m=0;m<DIM.height[comp];m++)
	frame[comp][m]=new pel_t [DIM.width[comp]];
    }

    //*******************************************************
    // get file handles (declared "static" to keep pointers to 
    // the files between each call to this function)
    //*******************************************************
    instream.open(infile,ios::in);//open file to read

  }
  //********************************************************
  //********************************************************

  //*******************
  // read in next frame
  //*******************
  if (kill)
    done=TRUE;
  else{
    for(comp=_Y_;comp<=_V_ && done==FALSE;comp++)
      for(m=0;m<DIM.height[comp] && done==FALSE;m++)
	for(n=0;n<DIM.width[comp] && done==FALSE;n++){
	  instream.get(ch);
	  if (instream.eof())
	    done=TRUE;
	  else
	    frame[comp][m][n]=(double)ch;
	}
  }

  if (done==FALSE){

    yuv_to_rgb(rgb_pels,&frame,DIM);

    //********************************************************
    // Release Java variables to allow Java to use the
    // new values assigned to them, and to avoid memory leaks
    //********************************************************
    env->ReleaseStringUTFChars(jinfile, infile);     
    env->ReleaseIntArrayElements(jpels, rgb_pels, 0);
    
  }
  else{

    instream.close();

    //kill 'frame'
    for(comp=_Y_;comp<_V_;comp++){
      for(m=0;m<DIM.height[comp];m++)
	delete frame[comp][m];
      delete frame[comp];
    }//for comp

    FrameNum=-1;

  }//done

  //********************************************************  
  //********************************************************
  
  FrameNum++;
  return(done);

}//Java_nat_YUVplay

